/**
* \file GameEngine.cpp
*
*/
#include "game_engine.hpp"
#include "IEngineConfig.hpp"
#include "config.hpp"
#include <fusion/core/CStringBuilder.hpp>
#include <fusion/core/string.hpp>
#include "audio/ISoundManager.hpp"
#include "config.hpp"
#include "entities/player.hpp"
#include <utf8.h>
#include "fx/CParticleManager.hpp"
//#include "audio/CFModExDriver.hpp"
#include "audio/ISoundManager.hpp"
//#include "audio/CNu.hpp"
#include "audio/IAudioDriver.hpp"
//#include "entities/IMultiModeTransformer.hpp"
#include "states/IGameState.hpp"
#include "states/COptionsState.hpp"
#include "IdList.hpp"
#include "entities/CWalkyrieYF19.hpp"
//#include "events/CInputManager.hpp"
#include "utils/debug.hpp"
#include <mirrpp/IPostProcessingHandler.hpp>
#include "utils/IGUIConsole.hpp"
#include  <xeffects/XEffects.h>

using namespace irr;
using namespace core;
using namespace video;
using namespace scene;
using namespace gui;
using namespace fus;

using namespace audio;


//fus::ELogLevel CGameEngine::_irrlichtLogLevelsMap[irr::ELL_NONE + 1];

//void
//updatePlayersDescriptors( const SIngameConfig::TMapping& mapping, CPlayer& player) {
//
//    //player.setDescriptors()
//    //BOOST_FOREACH()
//    for ( auto i = 0; i < NPlayerInput::Count;++i ){
//        player._mappingSet.addBind( static_cast<NPlayerInput::EId>(i), mapping[i]);
//    }
//}


CGameEngine::TParticleManager CGameEngine::_particleManager;
CGameEngine::TSpriteManager CGameEngine::_spriteManager;
//CGameEngine::TSoundManager CGameEngine::_soundManager;

//
CGameEngine::TBiMap CGameEngine::_videoDrivers;




// Function Init()
CGameEngine::CGameEngine() :
IGameEngine(),
_device( 0 ),   // nullptr
//_fpsCamera(0),
mLastTime(0),
_driver(0),
_smgr(0),
//_inputManager( new input::CInputManager()),
_console(0),
mElapsedTime(0),
_postProcessing(0),
IStateManager(),
//_postProcessingEffect(0),
// TEMPORAIRE:audio::CFModExDriver::instance()
_soundManager( 0 ),
_messageBoxesTab(0)
, _running(false)
, _yf19walkyrie(0)
//, _yf21walkyrie(0)
//_internalState(NEngineState::Uninitialized)
   // mRenderingMode(Fill)
{
    //std::fill(  .first(), .end(), 0);

    //createIrrlichtDevice();
    //registerGUIElementFactory
    EngineConfig.DriverType = conf::_DEFAULT_VIDEO_DRIVER;
}



void
CGameEngine::onStackEmptied()
{
    //!
    //TODO should abort program
    //"Stack emptied"
    Quit();
}



bool
CGameEngine::isCursorShown() const
{

    return _device->getCursorControl()->isVisible();
}





void
CGameEngine::setLogLevel( const fus::ELogLevel& level )
{

    Logger::reportingLevel() = level;
    //getLogger()->setLogLevel();
    _device->getLogger()->setLogLevel( convertFusionLogLevel( level ) );
}




fus::ELogLevel
convertIrrlichtLogLevel( const  irr::ELOG_LEVEL& level )
{

    switch(level)
    {


        case irr::ELL_WARNING:      return fus::ELogLevel::Warning;
        case irr::ELL_ERROR:        return fus::ELogLevel::Error;
        case irr::ELL_NONE:         return fus::ELogLevel::None;
        case irr::ELL_INFORMATION:
        default:                    return fus::ELogLevel::Notice;

    };
}





irr::ELOG_LEVEL
convertFusionLogLevel( const fus::ELogLevel& level )
{

    switch( level )
    {


        case fus::ELogLevel::Warning:      return irr::ELL_WARNING;
        case fus::ELogLevel::Error:        return irr::ELL_ERROR;
        case fus::ELogLevel::None:         return irr::ELL_NONE;
        case fus::ELogLevel::Notice:
        default:                           return irr::ELL_INFORMATION;
    };
}







// We could pass if loaded or not
void
CGameEngine::onNewTopMostState( IGameState* newState)
{
    //loaded
    irr::scene::ISceneManager* _smgr = newState->scene();

    //if( getTopMostState() )
    _device->setInputReceivingSceneManager( _smgr );

    showCursor( newState->ShowCursor );

    if( newState->getMainTab() != nullptr)
    {
        newState->getMainTab()->setVisible( true );
    }

}


//const TEngineParameters& engineParam,const SIngameConfig& gameConfig,const SAudioParameters& audioParam
// TODO renvoyer un int avec plusieurs codes d'erreurs ?
int
CGameEngine::Run(TGameState state)
{

//    if( _internalState == NEngineState::Uninitialized ){
//
//        // TODO alors ici on initialise avec les valeurs par defaut !!
//        //-supprimer les dependances XML, tout gerer avec lua
//        _LOG_ERROR << "You have to initialize the engine first";
//        return false;
//    }

    pushState(state);

    try
    {
        //createIrrlichtDevice(engineParam);


        this->synchronize();


        /// While irrlicht engine active and state available
        while( _device->run() && _running )
        {

            // State may have changed
            // We could detect if top most states hav changed
            this->synchronize();

            this->Update();
            this->Flush();
        }
        return true;
    }
    catch(std::bad_alloc& e){

    }
    catch(std::runtime_error& e){
        _LOG_ERROR  << "Runtime error" << e.what();
    }
    //std::invalid_argument
    //luabind::error
    //
    catch(std::exception& e){
        //exit()
        _LOG_ERROR << e.what();
    }
    catch(...){
        //_LOG_ERROR [ "Uncatched exception"  << "test" ]
        _LOG_ERROR << "Uncatched exception" ;
    }

    //TODO
    //this->CleanUp();

    return false;
};



//bool
//CGameEngine::switchConsoleState() {
//
//    if( switchBit(NBit::Console) ) {
//
//
//    }
//}
//




// helper function/wrapper for utf8
void
CGameEngine::popupMessage(const wchar_t* shortDescription,const std::string& longDescription){

    std::wstring temp;
    fus::convertIntoWideString(longDescription,temp);
    popupMessage(shortDescription, temp.c_str() );
}




void
CGameEngine::popupMessage(const wchar_t* caption,const wchar_t* longDescription)
{

    _internalState = NEngineState::Error;

    //_INFO <<  "msgBox tab children size before adding popup" << _messageBoxesTab->getChildren ().size();

    gui::IGUIWindow* window = gui()->addMessageBox(
                                        caption,            // title
                                        longDescription,    // description
                                        true,               // modal
                                        gui::EMBF_OK,        // type
                                        //EMBF_OK|EMBF_CANCEL
                                        _messageBoxesTab                //parent
                                        );

    //_INFO <<  "msgBox tab children size before addChild() " << _messageBoxesTab->getChildren ().size();

    // TODO checker si ca cree 2 enfants ou 1 seul
    _messageBoxesTab->addChild(window);
    _messageBoxesTab->setVisible(true);

    //_INFO <<  "msgBox tab children size after addChild" << _messageBoxesTab->getChildren ().size();


    // Put it in front of others
    if(!gui()->getRootGUIElement()->bringToFront(_messageBoxesTab) )
    {
        _LOG_WARNING << "Could not bring element to front";
    }

    //_INFO << "Popping up  message box [" << window << "]";
    //const wchar_t*
    //_LOG_WARNING << "Messagebox error message:" << longDescription ;

    this->showCursor(true);
}





//bool startScript(const char* scriptFilename);
void
CGameEngine::getAvailableVideoDriverList(std::vector<std::wstring>& list)
{

    list.clear();

    // let => int, right => wstring
    BOOST_FOREACH( TBiMap::right_const_reference p, _videoDrivers.right )
    {
        //p.second <<
//        _INFO << "test" << p.second;
        list.push_back(p.first);
    }


}



// TODO a corriger !!
//void
//CGameEngine::setCamPos(const float& x,const float& y,const float& z){
//
//    TVector3df vec(x,y,z);
//    getCamera()->setPosition( vec );
//}
//
//
//void
//CGameEngine::setCamTarget(const float& x,const float& y,const float& z){
//
//    TVector3df vec(x,y,z) ;
//    getCamera()->setTarget( vec );
//}



//irr::scene::ICameraSceneNode*
//switchFreeFlightCamera() {
//
//}

// Parcourt tous les elements et les autoscale
void
CGameEngine::enableGUIAutoScale(IGUIElement* element)
{

//    irr::gui::IGUIElement* root = _gui->getRootGUIElement ();
//    if(root){
//        root->setAlignment( EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE);
//    }
    //setRelativePositionProportional
    // Left / Right   : Top/Bottom
    //button->setAlignment( EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE);
}


void
CGameEngine::logDebugInformations()
{


    const irr::IOSOperator* os = _device->getOSOperator();

    _INFO << "Irrlicht version:\t" << _device->getVersion() << "\n"
           << "Type:\t" << _device->getType () << "\n"
           << "OS informations: " << os->getOperationSystemVersion()
           << "Driver info:\t" << _driver->getVendorInfo().c_str() << "\n"
           //COpenGLExtensionHandler
           << "Mipmap enabled" << _driver->queryFeature(EVDF_MIP_MAP )
           << "Liste des drivers supportes: Not available";

}


bool
CGameEngine::loadState(TGameState previousState,TGameState nextState)
{
//    FUS_ASSERT(state != 0);
    scene::ISceneManager* scene = nullptr;
    video::ITexture* savedScreenshot = nullptr;
//    core::dimension2du dims = driver()->getScreenSize() ;
//    dims.Width++;
//    dims.Height++;



    /**
    Recuperer le scenemgr du state juste en dessous plutot que d'utiliser une variable globale
    differences avec le gamestate d'avant ?
    TODO that shouldn't be here
    **/
    TTempAssociation preLoadElements;
    generateGuiList( CGameEngine::instance().gui()->getRootGUIElement(), preLoadElements );



    //!
//    if( state->CloneScene )
//    {



    //getTopMostState()->

    if(previousState)
    {
        //,ECF_A8R8G8B8

//        texture->grab();

//        _INFO << "Created RTT has size: " << driver()->getScreenSize().Width << "/" << driver()->getScreenSize().Height;
//        _INFO << "Created RTT has size: " << texture->getSize().Width << "/" << texture->getSize().Height;
//        _INFO << "Created RTT from driver: " << driver();
//


        printSceneManager( previousState->scene()->getRootSceneNode() );
        scene = previousState->scene()->createNewSceneManager( nextState->CloneScene );

        scene::ICameraSceneNode* cam = previousState->scene()->getActiveCamera();

        _INFO << "OLD ACTIVE CAMERA " << cam;
        _INFO << "NEW ACTIVE CAMERA " << scene->getActiveCamera();

        scene->setActiveCamera( cam );
        _INFO << "NEW ACTIVE CAMERA after setting old one" << scene->getActiveCamera();

//
//        //! TODO
        if(nextState->RenderSavedScreenshot)
        {
//
//            //addTexture (
            video::IImage* screenImg = driver()->createScreenShot(
//                                                          ERT_FRAME_BUFFER
                                                          );

            FUS_ASSERT(screenImg && "couldn't take screenshot");
            savedScreenshot = driver()->addTexture("screenshot", screenImg,0);

            FUS_ASSERT(savedScreenshot && "couldn't convert screenshot to texture");
            screenImg->drop();
//            savedScreenshot = nullptr;
////            nextState->SavedScreenshot = texture;
//            _INFO << "Taking screenshot of state " << previousState->getName();
//            previousState->takeScreenshot(texture);
//
//            getPostProcessing()->blurTexture(texture,texture);

        }


    }
    else
    {
        scene = _smgr->createNewSceneManager( nextState->CloneScene );
    }


//    scene->addCameraSceneNodeFPS();
//    scene->setActiveCamera();

//                #ifdef DEBUG
                    #if 0
                    TTempAssociation postLoadElements;
//                        std::list<irr::gui::IGUIElement*> result;
                    TTempAssociation result;
                    generateGuiList( CGameEngine::instance().gui()->getRootGUIElement(), postLoadElements);
                    compareLists(preLoadElements,postLoadElements, result);

                    // parcourir la liste
                    // boost::variant< ISceneNode*, IGUIElement* >
                    _LOG_WARNING << "Found " << result.size() << " differences ";
                    for(auto it(result.begin()), last(result.end()); it != last; ++it)
                    {
                        //!
                        //(*it)->getTypeName () <<
                        //<< (*it)->getID()
                        //if( it->first->getParent() != )
                        //isMyChild is recursive

                        // TODO describe gui element
                        gui::IGUIElement* tab = state->getMainTab();
                        if( tab && (tab->isMyChild(it->first) || (tab == it->first) ) )
                        {
                            _INFO << "That's ok, child from MainTab";
                        }
                        else
                        {
                            std::string temp;
                            fus::convertIntoNarrowString(it->first->getText(),temp);
                            _LOG_WARNING << "Unexpected element of type " << it->first->getTypeName()
                                        <<  "\nText: " << temp.c_str()
                                        <<  " with parent " << it->second;
                        }
                        //getText();
                    }


                #endif

    const bool result = nextState->load(this,  scene, savedScreenshot );

    if(savedScreenshot != nullptr){
        savedScreenshot->drop();
    }
    return result;

}





// TODO a revoir
// restartGraphicalDevice
void
CGameEngine::closeIrrlichtDevice()
{

    // If already exists
    if(_device)
    {


        // Sauvegarder la scene avant
        _device->closeDevice();
        _device->run();
        _device->drop();

        _driver = 0;
        _smgr   = 0;
        _gui = 0;

        //_fpsCamera->remove();
    }
}







bool
CGameEngine::setFont(const std::string& fontPath)
{

    IGUISkin* skin = _gui->getSkin();
    IGUIFont* font = _gui->getFont(fontPath.c_str());
    if (font){
        skin->setFont(font);
        return true;
    }
    else{
        skin->setFont(_gui->getBuiltInFont(), EGDF_TOOLTIP);
        return false;
    }

}




// time in ms
TTimer::TimeType
CGameEngine::getElapsedTime() const
{

    #ifdef _BC_DEBUG
        return 7;
    #endif

    if(getBoolean(NBit::BulletTime) )
    {
        //TODO a check
        return (mElapsedTime / 10.f);
    }
    return mElapsedTime;
};





// TODO: pouvoir passer le chemin o enregistrer l'image avec
void
CGameEngine::TakeScreenshot(const std::wstring& name)
{

    // TODO getTopMostState()->takeScreenshot()
    // TODO le sauver avec le nom du state
    IImage* img = _driver->createScreenShot();

    static int counter = 0;

    counter++;;
    // test =
    irr::io::path filename=fus::TWCharStringBuilder(name)(counter)(".jpg").str().c_str();

    _driver->writeImageToFile(img,filename);
}






/// \function Draw
/// Draw things commun to each state and then draw specific state
void
CGameEngine::Flush()
{



// TODO rajouter par dessus beginScene(false) apres le render du state pour ajouter la console par exemple ?

    //beginScene()
    driver()->setRenderTarget( ERT_FRAME_BUFFER,true,true);
//, true, true
//true, true
//erase depth and
//    driver()->beginScene();


    getTopMostState()->Draw();
    //_spriteManager->draw2Dsprites();


    driver()->setRenderTarget( ERT_FRAME_BUFFER, false, false);


    _console->draw();


    _debugTab->draw();

    _messageBoxesTab->draw();

    // flip buffers
    driver()->endScene();


}


bool
CGameEngine::eventToMainHandler(const SEvent& event)
{


    // Pass onto player eventsss
//    _player->OnEventForPlayer(event);

    //HandleEvents(){
    switch (event.EventType )
    {

        case EET_GUI_EVENT:
            {


            //s32 id = event.GUIEvent.Caller->getID();
            gui::IGUIElement* caller = event.GUIEvent.Caller;
            const gui::IGUIElement* element = event.GUIEvent.Element;
            //->getParent();

            switch(event.GUIEvent.EventType)
            {

/*
                case gui::EGET_ELEMENT_HOVERED:
                    //EGUIET_WINDOW
                    //EGUIET_MESSAGE_BOX
//                    if(caller->getType() == EGUIET_MESSAGE_BOX){
//
//                        _INFO << "messagebox hovered [" << caller << "]";
//                    }
                    //caller->getDebugName()
                    //_INFO << "Element hovered [" << caller << "] of type [" << caller->getType() << "messageBox [" << _messageBoxesTab << "]";
                    break;
*/
                case gui::EGET_ELEMENT_FOCUSED:
                    // EGUIET_MODALSCREEN ->getType()

                    //_INFO << "caller" << caller->getText();
//                    _INFO << "caller" << caller->getTypeName();

                    #ifdef DEBUG
                    {
                        std::string result;
                    convertIntoNarrowString( caller->getText(), result);

                    //_INFO << "caller" << caller;
                    _INFO << "caller text" << result;
                    //_INFO << "caller debug name" << caller->getDebugName ();
                    //_INFO << "caller debug name" << caller->getTypeName();
                    _INFO << "caller ID" << caller->getID ();
                    //_INFO << "caller type" << caller->getID ();
                    }

                    caller->setVisible(true);
                    _INFO  << caller->getTypeName();
                    #endif
//                    if(element){
//                        _INFO << "element" << element->getTypeName();
//                    }
                    break;

                case gui::EGET_EDITBOX_ENTER:
                    _INFO << "EVENT EGET_EDITBOX_ENTER RECEIVED";
                    /// verifier que ca vient de la console
                    if(caller->getID() == NGuiId::Console)
                    {

                        std::string temp;
                        // TODO ne pas flusher l'editbox avant
                        // distinguer commande de leur resultat ?
                        convertIntoNarrowString(caller->getText(),temp);

//                        _INFO << "Running command [" << temp << "]...\r\n";
//                        const bool ret = _vm.RunCommand( temp.c_str() );
//                        if(ret){
//                            dynamic_cast<IGUIConsole*>(caller)->flushCommand();
//                        }
//                        else {
//                            // TODO on genere une erreur
//                            // popupMessage
//                        }
                    }
                    break;

                case gui::EGET_MESSAGEBOX_OK:

                    _INFO << "Message box ok detected !";
                    //_INFO << "children nb" <<  _messageBoxesTab->getChildren ().size();
                    // TODO checker qu'il n'y a plus d'enfant
                    // if( _messageBoxesTab->isMyChild (caller)) {
                    if( (caller->getParent() == _messageBoxesTab) ){

                        //_INFO << "caller is child of messageBoxTab:";

                        caller->remove();


                        //if( _messageBoxesTab->getChildren ().size() == 1){
                        if( noActivePopup() )
                        {

                            _INFO << "MessageBox has no children left, setting it to invisible";
                            //
                            _internalState = NEngineState::Running;
                            _messageBoxesTab->setVisible(false);
                        }
                        return true;
                    }


                    break;



                default:
                    //_INFO << "GUI EVENT NOT HANDLED";
                    break;
            };
            }
            break;

        /// En cas de Log
//        case EET_LOG_TEXT_EVENT:
//            // TODO checker pk ca marche pas
//            //{
//                //const fus::ELogLevel level = convertIrrlichtLogLevel(event.LogEvent.Level);
//                _INFO << "IRRLICHT EVENT";
//                fus::ILogger::Log(convertIrrlichtLogLevel(event.LogEvent.Level)) << event.LogEvent.Text;
//            //}
//            return true;


        case EET_KEY_INPUT_EVENT :
            if(!event.KeyInput.PressedDown)
                return false;

            //if()

            // et checke si elle est pressee
            switch (event.KeyInput.Key)
            {

//                case irr::KEY_ESCAPE:
//                    //return true;

                case KEY_F1:
                    _INFO << "Switch console";
                    {
                        bool result = SwitchBit(NBit::Console);
                        _console->setVisible( result );
                        if(result){
                            gui()->getRootGUIElement()->bringToFront( _console );
                        }
                    }

                    return true;

                case KEY_F2:
                    _INFO << "Switch debug infos";
                    {

                        bool result = SwitchBit(NBit::DebugData);
                        _debugTab->setVisible( result );
                        getTopMostState()->onDebug( result  );
//                        if(result){
//                            gui()->getRootGUIElement()->bringToFront( _debugTab );
//                        }
                    }

                    return true;

//                case KEY_F2:
//
//                    _INFO << "Switch freeflight camera";
//                    // si on vient d'activer le mode free camera
//                    if(SwitchBit(NBit::FreeFlight)){
//
//                        // pauses game
//                        //SwitchBit(NBit::Pause,true);
//                        scene::ICameraSceneNode* cam = _smgr->setCamera();
//                        _cameras.push(cam);
//
//                        _smgr->setActiveCamera(_mainCamera);
//                        _mainCamera->setPosition(cam->getPosition());
//                        _mainCamera->setTarget(cam->getTarget());
//
//                    }
//                    else {
//                        // resumes game
//                        //SwitchBit(NBit::Pause,false);
//                        if(_cameras.size()){
//                            if(_cameras.top()){
//                                _smgr->setActiveCamera(_cameras.top());
//                                _cameras.pop();
//                            }
//                        }
//                    }
//                    break;

                case KEY_F3:
                    // Activate bullet time !
                    _INFO << "Switch bullet time";
                    SwitchBit(NBit::BulletTime);

                    return true;


                case KEY_F5:

                    _INFO << "switchin' to fullscreen mode (disabled)";
                    SwitchBit(NBit::Fullscreen);
                    //_INFO << "new fullscreen/test" << fullscreen << "/" << test;
                    return true;


                case KEY_F6:
                case KEY_F7:
                //
                    {


                    SwitchBit(NBit::DebugData);

                    }
                    return true;

                // TODO wireframe mode
                case KEY_F8:
                    {

                    _INFO << "switching between wireframe mode etC.. ";
                    //! TODO should loop through all materials of all devices ?
                    // or use globalMaterial ?

//                    const bool wireframe = SwitchBit(NBit::WireFrame);
//                    if(wireframe){
//                        //_smgr->getRootSceneNode()->setMaterialFlag(video::EMF_WIREFRAME,wireframe);
//                        _player->_node->setMaterialFlag(video::EMF_WIREFRAME,wireframe);
//                        //_player->_animatedNode->setMaterialFlag(video::EMF_POINTCLOUD, false);
//                    }
//                    else {
//                        //_smgr->getRootSceneNode()->setMaterialFlag(video::EMF_WIREFRAME,wireframe);
//                        _player->_node->setMaterialFlag(video::EMF_WIREFRAME,wireframe);
//                        //_smgr->getRootSceneNode()->setMaterialFlag(video::EMF_POINTCLOUD, true);
//                    }
                    }
                    return true;

                // TODO Afficher les hitboxes
                case KEY_F9:
                    //
                    _INFO << "switching display of hitboxes";
                    SwitchBit(NBit::DisplayHitCircles);
                    return true;

                // afficher les BB
                case KEY_F10:
                    return true;

                case KEY_F12:
                    TakeScreenshot(L"screenshot");
                    return true;


                default:
                    // Si jamais la console est activee

                    break;
            };

            break;



        default:
            break;
    }


    return false;
}



/*
 EGET_MESSAGEBOX_YES,
 EGET_MESSAGEBOX_NO,
 EGET_MESSAGEBOX_OK,
 EGET_MESSAGEBOX_CANCEL,
*/
/**
* Handle generic actions first ( that is common to every state ) and then state-specific events

**/
bool
CGameEngine::OnEvent(const SEvent& event)
{

    //_device->getTimer()->getTime()
//    if(_inputManager->handleEvent(  event)){
//        return true;
//    }


    if(event.EventType == EET_LOG_TEXT_EVENT )
    {


            fus::ELogLevel level  = fus::ELogLevel::Notice;
            switch(event.LogEvent.Level)
            {
                case irr::ELL_INFORMATION:  level = fus::ELogLevel::Notice; break;
                case irr::ELL_WARNING:      level = fus::ELogLevel::Warning; break;
                case irr::ELL_ERROR:        level = fus::ELogLevel::Error; break;
                case irr::ELL_NONE:         level = fus::ELogLevel::None; break;
                case irr::ELL_DEBUG:        level = fus::ELogLevel::Debug; break;
            };

    //        _INFO << "IRRLICHT EVENT";
            fus::Logger::Log(level) << event.LogEvent.Text;

            return true;
    }



//    // If console enabled
//    if(mBitSet[NBit::Console]){
//
//        return eventToConsoleHandler(event);
//
//    }
//    // else
//    else
//
    if(eventToMainHandler(event))
    {
        _INFO << "Event processed by main handler";
        return true;
    }
    //else if(!isStackEmpty()){

        // If there is an active state


        // let state handle it
        if(getTopMostState()->OnEvent(event))
        {
            _INFO << "Event processed by state";
            return true;
        }

    //}



    //let irrlicht handles event
    return false;
}//end of function




//template<typename T>
//irr::core::vector2d<T>
//CGameEngine::getAbsoluteDimension(const float& x,const float& y) const {
//
//    return irr::core::vector2d<T>(x*driver->getScreenSize().Width,y*driver->getScreenSize().Height);
//}

//getAbsoluterectangle(const irr::core::rect<f32>& rect){
irr::core::rect<s32>
CGameEngine::getAbsoluteRectangle(const float& x,const float& y, const float& x2,const float& y2) const{
    irr::core::rect<s32> tmp;
    tmp.UpperLeftCorner = getAbsoluteDimension<s32>(x,y);
    tmp.LowerRightCorner = getAbsoluteDimension<s32>(x2,y2);
    //tmp.UpperLeftCorner +
    return tmp;
}




irr::core::position2di
CGameEngine::getCenteredPosition(const irr::core::dimension2du& dimRect) const
{

//constrainTo
    const irr::core::dimension2du& screen = _driver->getScreenSize();
// getscreensize() renvoie un dimension2du
    irr::core::position2di pos;//(dimRect.Width);
    if( screen.Width < dimRect.Width ){
        pos.X = -(dimRect.Width -screen.Width);
    }
    else {
        pos.X = screen.Width - dimRect.Width;
    }

    if( screen.Height < dimRect.Height ){
        pos.Y = -(dimRect.Height -screen.Height);
    }
    else {
        pos.Y = screen.Height - dimRect.Height;
    }
   // temp.Y = (driver->getScreenSize().Height - std::max(screen.Height, dimRect.Height));
    pos /= 2;

    return pos;
}



bool
CGameEngine::addVideoDriverIfSupported( const std::wstring& name, const TVideoDriverId& id )
{

    if(IrrlichtDevice::isDriverSupported(id)){
        _videoDrivers.insert( TKeyAssociation(id,name) );
        return true;
    }
    return false;
}



//CConfigManager::initializeDictionaries
bool
CGameEngine::getVideoDriverFromName(const std::wstring& wname,TVideoDriverId& id)
{

    try {
        //TMappings::left_mconst_iterator _keyMap.left.find(keyCode);
        _videoDrivers.right.at( wname );
        return true;
    }
    catch( std::out_of_range & e ) {
        std::string name;

        convertIntoNarrowString(wname,name);
        _LOG_ERROR << "Couldn't find id associated with name [" <<  name << "] in keyCodeToString";
        return false;
    }
}



//void
//CConfigManager::addAudioDriver(){
//
//}


CGameEngine::~CGameEngine()
{
    // TODO ca aussi ca peut-tre un bug
    _device->drop();
}



//bool
//CGameEngine::loadConfig(const char* filename) {
//
//    _INFO << "TODO load config from lua scruipt";
//}
//





/// \function CleanUp()
void
CGameEngine::CleanUp()
{

    IStateManager::CleanUp();
}




/// \function Update()
void
CGameEngine::Update()
{



    ///////////////////////////////////////
    ////  Update timer
    ///////////////////////////////////////
    // TODO renommer now en currentTime

    const u32 now = _device->getTimer()->getTime();
    mElapsedTime = (now-mLastTime);
    mLastTime = now;

    // TODO process mElapsedTime to have sammler steps for example




    ///////////////////////////////////////
    ////  Update playerinput cache
    ///////////////////////////////////////
    GameConfig.MappingPlayer1.updateCache( _device->getInputManager(), CachedInputPlayer1 );

    //!
    if( _internalState != NEngineState::Running)
    {
        return;
    }

    ///////////////////////////////////////
    ////  Update sound
    ///////////////////////////////////////
    getSoundManager()->update();

    ///////////////////////////////////////
    ////  Should we show the cursor ?
    ///////////////////////////////////////
    bool shouldShowCursor = (!noActivePopup()) | _console->isVisible();
    if( shouldShowCursor == false){
        shouldShowCursor = getTopMostState()->ShowCursor;
    }

    if( shouldShowCursor ^ isCursorShown() ){
        this->showCursor( shouldShowCursor );
    }


    ///////////////////////////////////////
    ////  Treat state event queue
    ///////////////////////////////////////
    // fait dans Run
    //synchronize();

    ///////////////////////////////////////
    ////  Update window
    ///////////////////////////////////////
    #ifndef DEBUG
    //#error test
    if( _device->isWindowActive())
    {
    #endif

    // Won't update if on pause or console active
    // TODO that should update but with a time of 0 !
//        if(paused() || consoleEnabled() )
//            return;


        if( _debugTab->isVisible() )
        {
//            _debugTab->setText(
//                driver()->getPrimitiveCountDrawn ()
//                );
            // TODO mettre a jour la position

            std::wstring debugData;

            debugData = getTopMostState()->generateDebugData();

            fus::TWCharStringBuilder temp;
            temp(L"\nBluecosmos version: ")(conf::_GAME_VERSION )(L"\nCompiled at: ")( __DATE__ )(L"\nFPS: ")( driver()->getFPS() );
                temp(L"\nTriangles drawn: ")( driver()->getPrimitiveCountDrawn ( 1 ) )("\n\n")( debugData.c_str() );
                temp(L"\nElapsed Time: ")( mElapsedTime );

//getPrimitiveCountDrawn
//            std::wstring wtemp;

//            fus::convertIntoWideString( temp.str(), wtemp );
//            debugData.append( wtemp.c_str() );
//            debugData.append( L"\nBluecosmos version: ").append( conf::_GAME_VERSION )
////                        .append( L"\nLuabind version: ").append( LUABIND_VERSION )
//                        .append( L"\nCompiled at: ").append( __DATE__ )
//                        .append( L"\FPS:").append( driver()->getFPS() );

            _debugTab->setText( temp.str().c_str() );
        }
        // Otherwise update state
        getTopMostState()->onUpdate(mElapsedTime);


        //_spriteManager->update(getElapsedTime());
        getSoundManager()->update();

    #ifndef DEBUG
    }
    else
    {
        _device->yield();
    }
    #endif
}






/// \function Quit
/// Log why quitting ( user asked,error... ) and set running to false to break while
// Lui permettre d'interrompre tt type de script ( qui sont lancs alors en thraed ?
void
CGameEngine::Quit()
{

    _running = false;

    //Logger::Log()
    _LOG_WARNING << "User asked to leave game";
}



//irr::scene::ISceneManager*
//CGameEngine::scene()
//{
//    return _smgr;
//
////    if(isStackEmpty())
////    {
////        return
////    }
////
////    return getTopMostState()->scene();
//
//};



bool
CGameEngine::saveConfig( const SIngameConfig& gameConfig )
{
    _INFO << "Config saved";

    GameConfig = gameConfig;
    // TODO
    //saveConfigIntoLuaFile( configFilename , game.EngineConfig, GameConfig, game.AudioConfig );
    return true;
}




void
CGameEngine::pushOptionsState()
{
//    auto state = getTopMostState();
//    video::ITexture* texture = 0;
//
//    if(state)
//    {
////        texture = driver()->addRenderTargetTexture  ( driver()->getScreenSize() , state->getName().c_str() );
//        state->takeScreenshot(texture);
//
//    }

    pushState( new COptionState( GameConfig) );
}





//const TEngineParameters& engineParam,const SIngameConfig&,const SAudioParameters&
bool
CGameEngine::InitEngine(const SIngameConfig& gameConfig)
{
    GameConfig = gameConfig;

    // Ces 3 la suffisent
//    addVideoDriverIfSupported(L"OpenGL",EDT_OPENGL);
//    addVideoDriverIfSupported(L"Direct3D9",EDT_DIRECT3D9);
//    addVideoDriverIfSupported(L"Direct3D9",EDT_DIRECT3D8);

//    // Try to
//    //_LOG_DEBUG << "get irrlicht parameters";
//    config->getIrrlichtParameters(parameters);
//
//    if(!getVideoDriverFromName(config->_videoDriver, parameters.DriverType) || (!irr::IrrlichtDevice::isDriverSupported(parameters.DriverType)) ){
//
//        _LOG_WARNING << "driver [" << parameters.DriverType << "] unsupported";
//        parameters.DriverType = conf::_DEFAULT_VIDEO_DRIVER;
//
//    }
//

    //_INFO << "Using video driver [" << engineParam.DriverType << "] ";



    /////////////////////////////////////
    ///// Initialisation d'irrlicht
    /////////////////////////////////////
    //_LOG_DEBUG << "create irrlicht device";
    if( !createIrrlichtDevice( ) )
    {
        throw std::runtime_error("Couldn't create irrlicht device");
        //"Could not create device"
        //_LOG_ERROR << "Couldn't create irrlicht device";
        return false;
    }


    /////////////////////////////////////
    ///// Soundmanager Initialisation
    /////////////////////////////////////
    // -- Init audio manager
    //_INFO << "Audio driver" << config->_audioDriver;
    _soundManager = createSoundManager(AudioConfig);

    //_soundManager.reset( new CSoundManager(AudioConfig) );
    //if( !getSoundManager()->init() ){
    if( !_soundManager )
    {
        _LOG_ERROR << "problem with audio initialization";
    }



    /////////////////////////////////////
    ///// Initialisation of message dispatcher
    /////////////////////////////////////

    // TODO should go elsewhere
//    if( !bindEverythingToLua( ) )
//    {
//        _LOG_WARNING << "Could not initialize lua";
//        this->Quit();
//        return false;
//    }



    // create Player
    // TODO creer le joueur quand on en a besoin ?
//    if(!createPlayer())
//    {
//        _LOG_WARNING << "Could not create player";
//    }


    ///////////////////////////////////////
    ////  Setup postprocessing
    ///////////////////////////////////////
    _postProcessing = new IPostProcessingHandler();

    core::stringc msg;
    if ( ! _postProcessing->init( driver(), msg ) )
    {

        //
        _LOG_ERROR << msg.c_str();
    }
    else {
        _INFO << msg.c_str();
    }


    // TODO garder que un des 2
    _internalState = NEngineState::Running;
    _running =true;


    if( !driver()->queryFeature(video::EVDF_TEXTURE_NSQUARE)){
        _LOG_WARNING << "Non-squared textures not supported";
    }


    if( !driver()->queryFeature(video::EVDF_TEXTURE_NPOT)){
        _LOG_WARNING << "Non-squared textures not supported";
    }

    _INFO << "Engine created";



    return true;
}



bool
CGameEngine::enableJoysticks()
{

    return false;
}


const char* DriverNames[] = {
"null",
"software",
"EDT_BURNINGSVIDEO",
"D3D8",
"D3D9",
"OGL",
"Count"

};


//const irr::video::E_DRIVER_TYPE& driverType,blo::tribool fullscreen,blo::tribool vsync
//TEngineParameters
//TEngineParameters& parameters
bool
CGameEngine::createIrrlichtDevice()
{

    TEngineParameters& parameters = EngineConfig;


    _INFO << "using video driver [" << DriverNames[ parameters.DriverType ]  << "]";


    //BOOST_ASSERT( irr::IrrlichtDevice::isDriverSupported(parameters.DriverType) );
    if( !irr::IrrlichtDevice::isDriverSupported(parameters.DriverType) )
    {

        // On passe en BurningVideoSoftware
        parameters.DriverType = EDT_BURNINGSVIDEO;
        _LOG_WARNING << "Driver not supported, trying with BURNINGSVIDEO software driver";
        throw std::runtime_error("Driver not supported");
    }

    bool restart = false;


    // If device exists and active
    //&& device->run()
    if(_device )
    {
        //cleanIrrlichtTiedMembers();
        restart = true;
        closeIrrlichtDevice();
    }


    _device = createDeviceEx(parameters);

    if (!_device)
    {
        //_LOG_ERROR << "Could not create device";

        throw std::runtime_error("Could not create device");
        return false;
    }



    // run device once
    _device->run();

    _device->setWindowCaption( conf::_WINDOW_TITLE);
    // Necessaire sinon plante


    //var public
    _driver = _device->getVideoDriver();
    _smgr   = _device->getSceneManager();
    _gui    = _device->getGUIEnvironment();

    _INFO << "Device created with driver " << driver()->getName();

    // TODO supprimer ce genre de betise ca fait tout foirer
    //recharger la scene
    if(restart)
    {

    }


    irr::core::recti rect(0,0,500,200);


    ///////////////////////////////////////////////////
    ////    Create free flight camera
    ///////////////////////////////////////////////////


    /////////////////////////////////////
    /////  Setup event reciever
    /////////////////////////////////////
//    CEventManager::setDevice( _device );
    //mEventManager.reset( new CEventManager() );
    //mEventManager.reset( new CInputManager() );

    _device->setEventReceiver( this );
    //mEventManager->AddEventReceiver( this );

    // LogLevel defini dans les parameters normalement
    //_device->getLogger()->setLogLevel( ELL_WARNING ); //EALL_INFO


    ///////////////////////////////////////////////////
    ////    Create console
    ///////////////////////////////////////////////////
    _console = new irr::gui::IGUIConsole(
                                        gui(),
                                        core::recti(0,0, getScreenSize().Width,conf::_CONSOLE_HEIGHT),
                                        gui()->getRootGUIElement(),
                                        NGuiId::Console
                                        );

    _console->setVisible(false);
    _console->setBackgroundColor( video::SColor(100,100,100,100));

    _console->addButton(L"Test",L"print('Test reussi')");





    ///////////////////////////////////////////////////
    ////    Create debug data tab
    ///////////////////////////////////////////////////

//addStaticText (const wchar_t *text, const core::rect< s32 > &rectangle, bool border=false, bool wordWrap=true, IGUIElement *parent=0, s32 id=-1, bool fillBackground=false)=0

    _debugTab = gui()->addStaticText (L"temp"
                                     , core::recti(0,0, 300, getScreenSize().Height )
                                     , false    //!< border
                                     , true     //!< wordWrap
                                     , 0        //!< IGUIElement *parent=
                                     , NGuiId::DebugData
                                     , false      //!< fillBackground
                                     );


    _debugTab->setVisible(false);
    _debugTab->setBackgroundColor( video::SColor(100,100,100,100));




    // Setup override material
//    SOverrideMaterial& material = driver->getOverrideMaterial();
//    material.Enabled = true;
//    material.EnableFlags = irr::video::EMF_LIGHTING;
//    material.Material.Lighting = false;
//    //setFlag(irr::video::EMF_LIGHTING, false);
//    //E_SCENE_NODE_RENDER_PASS
//    material.EnablePasses = irr::scene::ESNRP_SOLID;
    /////////////////////////////////////
    /////  Setup backup Camera debugCamera
    /////////////////////////////////////
    _fpsCamera = _smgr->addCameraSceneNodeFPS(_smgr->getRootSceneNode(),100.0f,0.05f);
    // changer la vitesse de la camera


    /////////////////////////////////////
    /////  Particle manager
    /////////////////////////////////////
    // #define MessageDispatcher = this
    //this->addListener(this, NMessageType::CreateExplosion);
    //this->addListener(this, NMessageType::CreateExplosion);



    /////////////////////////////////////
    /////  SPrite manager
    /////////////////////////////////////
    _spriteManager.reset( new CSpriteManager( device() ) );



    /////////////////////////////////////
    /////  Postprocessing manager
    /////////////////////////////////////
//    _postProcessing.reset( new CSpriteManager( device() ) );



    /////////////////////////////////////
    /////  Particle manager
    /////////////////////////////////////
//    _particleManager.reset( new CParticleManager( device() ) );

    /////////////////////////////////////
    /////  Message box setup
    /////////////////////////////////////
    //
//    _messageBoxesTab = gui()->addMessageBox(L"Caption",L"No message",true,gui::EMBF_OK);
// TODO recuperer la surface de l'ecran
    _messageBoxesTab = gui()->addTab( core::recti(0,0,0,0), 0, NGuiId::ErrorTab );
    _messageBoxesTab->setVisible(false);
    // set whole screen
    _messageBoxesTab->setRelativePositionProportional( core::rectf(0.0f,0.0f,1.f,1.f));


    /////////////////////////////////////
    /////  load walkyrie
    /////////////////////////////////////
    _yf19walkyrie = new CWalkyrieYF19();

    scene::IAnimatedMeshSceneNode *node = _smgr->addAnimatedMeshSceneNode(
//                                                _smgr->getMesh(conf::_YF19_MESH_FILENAME)
                                                _smgr->getMesh(conf::_YF21_MESH_FILENAME)
                                                , 0             //!< Parent
                                                , NSceneId::YF19Walkyrie
                                            );

    #ifdef DEBUG
        node->setName("YF19");
    #endif
    node->getMaterial(0).Lighting = false;

    //NWalkyrieModes::
    _yf19walkyrie->load( node ,
                        IWalkyrie::FighterMode

//                            Modes[static_cast<int>(EWalkyrieMode::Fighter) ]

                         );

//    if(! _yf19walkyrie->load( node , EWalkyrieMode::Gerwalk ) )
//    {
//        //!
//    }

//    CGameEngine::getInstance().setup( _yf19walkyrie , _mappingSet );


    // TODO a deplacer
    setFont("images/hud.xml");
    return true;
}


//
//void
//CGameEngine::updateCacheMappingForPlayer1(std::map< NPlayerInput::EId, SBindCache >& cachedInput ) const
//{
//    //!
//    GameConfig.MappingPlayer1.updateCache( _device->getTimer()->getTime(), cachedInput );
//}



//const NWalkyrieModel::EId& id
IWalkyrie*
CGameEngine::getWalkyrie( const NWalkyrieModel::EId& id )
{
//    return  _walkyries[id];
    return  _yf19walkyrie;


//    else
//    {
//
//        setTransformerModel( conf::_YF21_MESH_FILENAME, conf::_YF21_BOOSTER_LEFT_NAME, conf::_YF21_BOOSTER_RIGHT_NAME);
//    }
//
//// Clear Mode aller chercher le mode actuel
//    clearModes(NWalkyrieModes::Fighter);
//
//    allowMode(NWalkyrieModes::Fighter );
//    allowMode(NWalkyrieModes::Gerwalk);
//    allowMode(NWalkyrieModes::Battloid);
//
//
//    // -- can always transform
//    _walkyrie->enableCyclicTransformation(true);
}
