/*! \file main.cpp
 *   \brief Here is the main loop of the game
 *
 *   The engine is first instanced in this file and then all methods
 *   are called successively inside the while() structure
*/
#include "IEngineConfig.hpp"
#include "game_engine.hpp"
#include "def.hpp"
#include "startup_config.hpp"
#include "config.hpp"
#include "entities/IWalkyrie.hpp"
#include "states/CTransitionState.hpp"
#include "states/CMenuState.hpp"
#include "states/CDummyPlayState.hpp"
#include "states/IPlayStateLua.hpp"
#include <fusion/logging/CConsoleLogger.hpp>
#include <fusion/logging/CFileLogger.hpp>
#include <fusion/scripting/CLuaVM.hpp>

//#include <boost/detail/lightweight_test.hpp>
#include <boost/smart_ptr.hpp>
//#include "events/CInputManager.hpp"



//! The main function
int
main (int argc, char *argv[])
{

    //! Begins logging every event to help debugging the game ( yes there are bugs ! )
    // On doit pouvoir passer des streams
    #ifndef _DEBUG
    fus::Logger::addLogger(new fus::CConsoleLogger());
    #endif

    fus::Logger::addLogger(new fus::CFileLogger("log.txt"));

    //  typedef void (*terminate_handler) ();
//    terminate_handler error_handler;
//    std::set_terminate( error_handler );
    //std::set_unexpected

    fus::Logger::reportingLevel() = fus::ELogLevel::Notice;
    //! Instanciation of the game engine

    // TODO appels des fct statique

    initializeKeycodes();
    initializeDeviceCodes();
    //CEventManager::initStaticMembers();
    IWalkyrie::initialize();

    fus::CLuaVM::setMessageHandler(&fus::onLuaError);

    // TODO we should e able to parse command linea bit more
    // If more than one param than next param is filename of config to load
    if(argc > 1)
    {

    }

//    try {
    //! TODO config par defaut:
    SIngameConfig GameConfig;
    GameConfig.EnableDialogs = true;
    //! TODO setup Player mapping



    CGameEngine& game( CGameEngine::instance() );
//    }
//    catch(...){
//        _LOG_ERROR << "test";
//    }
//    IMultiModeTransformer::initialize();


//    const char* filename = "config.lua";
    const char* configFilename = conf::CONFIG_FILENAME;

    // TODO nom du fichier = 1er param de la ligne de commande normalement argv[1]
    try
    {

        _INFO << "Loading config file with default driver type [" << game.EngineConfig.DriverType << "]";
        loadEngineConfigFromLua( configFilename , game.EngineConfig, GameConfig, game.AudioConfig );
    }
    catch(...)
    {
        _LOG_ERROR << "Could not load config [" << configFilename << "]";

        try {
            saveConfigIntoLuaFile( configFilename , game.EngineConfig, GameConfig, game.AudioConfig );
        }
        catch(...) {
//        if( !saveConfigIntoLuaFile("config.lua", game.EngineConfig, game.GameConfig, game.AudioConfig )) {
//            _LOG_ERROR << "Could not save config";
//        }
            _LOG_ERROR << "Could not save config";

        }
    }


    game.EngineConfig.DriverType = video::EDT_OPENGL;
//    game.EngineConfig.DriverType = video::EDT_DIRECT3D9;


    // TODO use a CNullDriver
    game.InitEngine( GameConfig );



//    return game.Run( new CMenuState() );
//    return game.Run( new CDummyPlayState() );

// 1er p
    return game.Run( new CLuaPlayState(nullptr,"levels/tutorial.lua") );

   // return EXIT_SUCCESS;            //!< EXIT_SUCCESS = 0
};
