/**
* \file single.cpp
*/
#include "CDummyPlayState.hpp"
#include <cmath>
#include <irrlicht.h>
#include "../entities/player.hpp"
#include "../entities/CFighterAnimator.hpp"
#include "../entities/IWalkyrie.hpp"
#include "../config.hpp"
#include "../game_engine.hpp"
#include "../states/COptionsState.hpp"
#include "../states/CScoreState.hpp"
#include "../states/CErrorState.hpp"
#include "../states/CDialogStatePlain.hpp"
#include "../IdList.hpp"
#include "../fx/CParticleManager.hpp"
#include "../utils/CBitSet.hpp"
#include "../entities/IMultiModeTransformer.hpp"
#include "../fleets/CAlliedFleet.hpp"
#include "../fleets/CEnemyFleet.hpp"
#include <fusion/maths/macros.hpp>
#include <fusion/core/string.hpp>



#define MOVE_FUNCTION(name) TPlayerMoveFunction CDummyPlayState::(name)(

const int _OSCILLATION_NUMBER = 3;


using namespace irr;
using namespace video;
using namespace core;
using namespace scene;


DECLARE_GAME_SINGLETON( CDummyPlayState )






// Passer le nom du script en parametre ?

//! Constructor
CDummyPlayState::CDummyPlayState() :
IPlayableState("Dummy play state")
, _camera(0)
//, _world(0)
{
    //(irr::scene::ISceneNode*)
    ADD_REQUIRED( _camera ,"main camera" );
//    ADD_REQUIRED( _world );
};


//! Destructor
CDummyPlayState::~CDummyPlayState()
{
    //_world->remove();
//    delete _script;
}




bool CDummyPlayState::load(CGameEngine* engine,irr::scene::ISceneManager* smgr,irr::video::ITexture* screenshot)
{

    //! should get it in constructor ?
//    CPlayer& player = getPlayer(0);
    if (IPlayableState::load( engine,smgr,screenshot ) == false)  return false;

    const core::vector3df pos(0,10,0);
    const core::vector3df target(0,0,0);

    //!bool cloneContent=false
    //engine().scene()->createNewSceneManager(false);

    //scene::ISceneManager* smgr = scene();

    _camera = smgr->addCameraSceneNode (
                0       //!< parent
                , pos   //!< position
                , target    //!< Target
//  s32  id = -1,
//  bool  makeActive = true
        );



    smgr->getRootSceneNode()->addChild( getPlayer(0).getNode() );

    // leur donner des IDS
//    ISceneNode* skydome = smgr->addSkyDomeSceneNode( engine().driver()->getTexture("images/skydome.jpg")
////                              , u32 horiRes=16, u32 vertRes=8, f32 texturePercentage=0.9, f32 spherePercentage=2.0, f32 radius=1000.f, ISceneNode *parent=0, s32 id=-1)
//                              );


    // add a nice skybox
    video::IVideoDriver* driver = engine->driver();
    driver->setTextureCreationFlag(video::ETCF_CREATE_MIP_MAPS, false);

// this is tricky, values after # or ## are not expanded by preprocessor, that's why we need another macro that will expand the other
//    #define STRINGIFY( text )  #text
    #define BUILD_PATH(suffix)   "images/irrSkybox/irrlicht2_" #suffix ".jpg"
    #define GET_TEXTURE(suffix) driver->getTexture( "images/irrSkybox/irrlicht2_" #suffix ".jpg" )
//    #define GET_TEXTURE(suffix) driver->getTexture( BUILD_PATH(suffix) )
    ISceneNode* skybox = smgr->addSkyBoxSceneNode(
            GET_TEXTURE(up),
            GET_TEXTURE(dn),
            GET_TEXTURE(lf),
            GET_TEXTURE(rt),
            GET_TEXTURE(ft),
            GET_TEXTURE(bk)

                        );
#undef GET_TEXTURE
#undef BUILD_PATH


    engine->driver()->setTextureCreationFlag(video::ETCF_CREATE_MIP_MAPS, true);

    CFighterAnimator* fighterAnimator = new CFighterAnimator();
    IWalkyrie::FighterMode->setDefaultAnimator( fighterAnimator  ) ;
    IWalkyrie::GerwalkMode->setDefaultAnimator( fighterAnimator  ) ;
    IWalkyrie::TransformingMode->setDefaultAnimator( fighterAnimator  ) ;

    _playerPreviousPos = Player.getPosition();

    _camera->setPosition( getPlayer(0).getWalkyrie()->getCurrentMode()->getCameraRecul() );
//_INFO << "loading";

    #ifdef DEBUG
        _camera->setName("Camera");
//        skydome->setName("Skydome");
        skybox->setName("Skybox");
    #endif

    return true;
}



void
CDummyPlayState::unload()
{
//    IGameState::unload();
}




void
CDummyPlayState::onDebug(const bool& on)
{


    IAnimatedMeshSceneNode* node = getPlayer(0).getWalkyrie()->getNode();
    FUS_ASSERT(node);

    if(on)
    {
        node->setDebugDataVisible( EDS_SKELETON );


        SColor c ( 255,255,255,255);

        SMaterial& Material = AccessibleArea.Material;
        Material.MaterialType = EMT_SOLID;
        Material.Wireframe = false;
        Material.Lighting = false;
        Material.ZWriteEnable = false;
        Material.BackfaceCulling = false;
        AccessibleArea.Vertices.reallocate(4);

        AccessibleArea.Indices.reallocate(6);
        AccessibleArea.Vertices[0] = S3DVertex(conf::_X_MIN,0,conf::_Z_MAX, 0,1,0, c , 0,0);
        AccessibleArea.Vertices[3] = S3DVertex(conf::_X_MIN,0,conf::_Z_MIN, 0,1,0, c , 0,1);
        AccessibleArea.Vertices[2] = S3DVertex(conf::_X_MAX,0,conf::_Z_MIN, 0,1,0, c , 1,1);
        AccessibleArea.Vertices[1] = S3DVertex(conf::_X_MAX,0,conf::_Z_MAX, 0,1,0, c , 1,0);



        AccessibleArea.Indices[0] = 0;
        AccessibleArea.Indices[1] = 1;
        AccessibleArea.Indices[2] = 3;
        AccessibleArea.Indices[3] = 1;
        AccessibleArea.Indices[4] = 2;
        AccessibleArea.Indices[5] = 3;

    }
    else {
        node->setDebugDataVisible( EDS_OFF );
    }
};

//
//void
//CDummyPlayState::onDebugDisactivation() {
//
//};

void
CDummyPlayState::renderDebugData()
{

    _INFO << "Drawing mesh buffer";
    driver()->setTransform(video::ETS_PROJECTION, core::IdentityMatrix);
    driver()->setTransform(video::ETS_VIEW, core::IdentityMatrix);
    driver()->setTransform(video::ETS_WORLD, core::IdentityMatrix);
    driver()->drawMeshBuffer( &AccessibleArea );
}









//
//
//void
//CDummyPlayState::processMovesAsGerwalk(
//                    DECLARE_PLAYER_MOVEFUNC_PARAMS
//          )
//
//{
//    // does nthg
//}
//







void
CDummyPlayState::processShootings( CPlayer& player, const SBindCache& mainShoot,const SBindCache& secondShoot )
{

//    const bool mainShoot = _cachedInput[ FULL_ID(NPlayerInput::MainShoot,0) ].pressed();

 //&& (_elapsedTimes[SinceLastShot] > 10 )
//
//    if( mainShoot.pressed()  )
//    {
//
//        // selon le modeshoot()
//        _INFO << "mainShoot : cooling down" << coolingDown();
//        FirePrimaryShot();
//
//        //bind->isDown()
//
//    }

/// ENORME HACK !! TODO faut mettre un repeat de 0
#if 0
    if( secondShoot )
    {

        _INFO << "Missile bind activated";

        // Means key was already pressed the precedent frame
        if( !isTrackingEnemies() )
        {

            //
            startTrackingEnemies();
        }
        // Player starts targeting enemies
        else
        {
            const bool effect = trackNewBattery();
            // If no battery left to aim at, stop firing missiles
            if(!effect)
            {
                stopTrackingEnemies();
            }

        }
    }
    else if( isTrackingEnemies() )
    {
        stopTrackingEnemies();
    }

#endif
}


/// Function Update()
void
CDummyPlayState::onUpdate(const TTimeUnit& elapsedTime)
{
//    _INFO << "onupdate";
    updatePlayer( 0
                 ,elapsedTime
//                 , engine().getCachedInput()
                 ) ;

//    _INFO << "after player update";
//    player.setPosition( )

    getAlliedFleet()->update( elapsedTime );


    // trying to update camera
    // TODO camera animator qui prend le milieu de target et previous position

    // Ca c quand on est en single player
    //! we always look a bit ahead of the player

    TVector3df playerPos = Player.getPosition() ;
    TVector3df previousPlayerPos = _playerPreviousPos;
    TVector3df oldCamPos = _camera->getPosition();
    TVector3df targetedCamPos = getPlayer(0).getWalkyrie()->getCurrentMode()->getCameraRecul() ;

//    TVector3df relPlayerMove = playerPos - previousPlayerPos;
//    if()

    //! we always look a bit ahead of the player
    TVector3df newTarget(playerPos + TVector3df(0,0,5));
    float xRange = ( conf::_X_MAX - conf::_X_MIN);

    newTarget.X = core::clamp(newTarget.X, conf::_X_MIN + xRange/3.f , conf::_X_MAX - xRange/3.f);
    _camera->setTarget( playerPos + TVector3df(0,0,5) );


    // if player goes back , camera goes to where it should be

    targetedCamPos.Z += Player.getPosition().Z;

    //! on prend le milieu entre les 2
    TVector3df newCamPos = oldCamPos + (targetedCamPos - oldCamPos)/4.f;
//    if( newCamPos.getLengthSQ () < 0.1)
//    {
//        newCamPos = targetedCamPos;
//    }

    _camera->setPosition( newCamPos );


//    oldCamPos.Y += 0.5;
//    _camera->setPosition( oldCamPos );

}





float
CDummyPlayState::getProgression() const
{
    // TODO
    return 0.f;
}




#if 0

// TODO a deplacer dans le gestionnaire de camera

/**
TODO creer un animator pour ca
**/
void
CDummyPlayState::updateCamera()
{

    TVector3df posCamera = getPlayer(0).getWalkyrie()->getCameraRecul();


    if(_shakingTimer)
    {

        _shakingTimer->update( engine().getElapsedTime() );
        if( _shakingTimer->reached() ){
        //if(_shakingTimer->update( engine().getElapsedTime() )){
            //cam->getTarget() + core::vector3df(0,-4,+20)
            //_CAM_POS
            //_camera->setPosition( _camPosBackup );
            delete _shakingTimer;
            _shakingTimer = 0;
            _LOG_DEBUG << "END SHAKING";
        }
        else {
//1.f/_shakingTimer->getElapsedTime()
            double ratio = _shakingTimer->getAdvancement();
            // Le mettre
            double intensite = 20.f*_FUSION_POW2(1.f-ratio);
            //float coeff = _SHAKING_LENGTH/(_OSCILLATION_NUMBER*M_PI);
            // Le ratio est cense etre compris entre 0 et 1

            //getElapsedTime()/_SHAKING_LENGTH;
            intensite *= std::sin( 4* PIperso * ratio );
            _INFO << "shaking d'intensite/ratio " << intensite << "/" << ratio;

            TVector3df decalage(intensite,0,0);


            //_camera->setPosition(_camPosBackup + decalage);
            posCamera += decalage;
            //cam->setPosition();
        }
    }
    // if player transforming
    //else

    /////////////////////////////////////////////////////
    //////  Debug des specs
    /////////////////////////////////////////////////////

}



void
CDummyPlayState::shakeCamera()
{

    _INFO << "shaking camera";
    // If already shaking
    if(_shakingTimer){
        _shakingTimer->restart();
    }
    else {
        _shakingTimer = new TTimer( conf::_SHAKING_LENGTH);
        //
        _camPosBackup = _camera->getPosition();
    }

}

#endif




void
CDummyPlayState::onResume()
{
    //_levelMgr.newGame();
    //getLevelManager()._playState = CGraphicalLevel::Playing;
    //_player->setupForNewGame();
}

//void  CDummyPlayState::onHold() ;









// TODO deplacer ca dans lua
//const float& playerAdv,
//, CDummyPlayState* _state
/*
void
CDummyPlayState::update(const TTimer::TimeType& elapsedTime){

//    _layer->getMaterial(0).getTextureMatrix(0).setTextureTranslate(0.f,-percievedSpeed);

    speed = _player->_walkyrie->getSpeed(SModeDescription::DefaultForward)/100.f;

    // plus on est eloigne du bas de l'ecran plus la sensation de vitesse
    float ecart = (_player->getPosition().Z- conf::_LOWER_LIMIT);
    ecart /= (conf::_UPPER_PART - conf::_LOWER_LIMIT);
    // Comme ca on a un ecart < 1

    //_INFO << "ecart " << ecart;
    speed += ecart/1000.f;// /10.f


    //_INFO << "upper progress" << _levelMgr._upperProgress/100.f;

    percievedSpeed += (speed);

    // TOOOOOOODOOOO : a retirer !!
    //BUG
    getLevelManager().updatePlayerProgress(percievedSpeed);


    //_INFO << "speed" << translation;

    // Selon l'avancement du joueur, incliner plus ou moins la camera
    this->updateCamera();


}
*/


////setUp cameras
//void
//CDummyPlayState::setupCameras(){
//
//    //irr::core::dimension2du dim =  engine().driver->getScreenSize();
//    //_cameras[Scene] =
//    _camera = scene()->addCameraSceneNode();
//                    /*0,  //!< parent
//                    //_CAM_POS, //!< cam pos
//                    //core::vector3df(0, 0, 0)    //!< lookat
//
//                    );*/
//
//    _camera->setPosition( _player->_walkyrie->_cameraRecul );
//
//    scene()->setActiveCamera(_camera);
//
//}



//! Render a frame
// en gros l'ecran a une taille de 60
void
CDummyPlayState::Draw()
{

    _INFO << "starting drawing CDummyPlayState";


    IPlayableState::Draw();


}






void
CDummyPlayState::onHold()
{


}






/// \function HandleEvents
bool
CDummyPlayState::OnEvent(const irr::SEvent& event)
{

    //! if already handled
    if(handleMainEvents(event))     return true;

    //! According to event type
    switch(event.EventType)
    {

        // Handle keys
        case EET_KEY_INPUT_EVENT :

            // check that's no key up
            if(!event.KeyInput.PressedDown)
                return false;

            switch (event.KeyInput.Key)
            {

                case irr::KEY_ESCAPE:
                    return true;



                //! Temp to test creating dialogs
                case irr::KEY_KEY_D:
                    //
                    {

                        _INFO << "Starting new dialog !!";
//                        CDialogStatePlain* dlg = new CDialogStatePlain();
//                        engine().pushState( dlg );
//                        CDialogStatePlain* dlg = new CDialogStatePlain();
//                        engine().pushState( dlg );

//                        //TPlayerReply
//                        dlg->addAnswer("Answer 1",0);
//                        dlg->addAnswer("Answer 2",0);
//                        dlg->addAnswer("La réponse 42",0);

                    }
                    return true;
                    //break;


                // Cree une explosion
                // sous forme de sprite
                // Activate editor
                case irr::KEY_KEY_E:


//                    _INFO << "add sprite explosion (disabled)";
                    //_levelMgr.addSpriteExplosion(irr::core::vector3df(5,0,0));

                    break;


                // cree une explosion avec spark
//                case irr::KEY_KEY_G:
//                    {
//                    //boost::any pos;
//                    //engine().addMessage(NMessageType::CreateExplosion,pos);
//
//                    //_levelMgr.addExplosion(irr::core::vector3df(5,0,0));
//                        _INFO << "Add spark explosion";
//                    //engine()._particleManager
//                        //ISceneNode* node = CGameEngine::getParticleManager()->createExplosion();
//                        //node->setParent( scene()->getRootSceneNode() );
//                        addSimpleExplosion( TVector3df(0,0,0) );
//                        this->shakeCamera();
//                    }
//                    break;
//
//                case irr::KEY_KEY_N:
//
//                    shakeCamera();
//                    break;

                // Tue le joueur
                case irr::KEY_KEY_M:

                    _INFO << "Mort du joueur";
//                    getPlayer(0).kill();
//                    getLevelManager().updateState();
                    break;

                default:
                    break;
            }
            break;

        default:
            break;
    } // end 1nd switch





    return false;
}



